@@ -3,6 +3,7 @@ Automated Time Series Model Selection for Python
 
 https://github.com/winedarksea/AutoTS
 """
+
 from autots.datasets import (
     load_hourly,
     load_daily,
@@ -19,14 +20,14 @@ from autots.datasets import (
 from autots.evaluator.auto_ts import AutoTS
 from autots.evaluator.event_forecasting import EventRiskForecast
 from autots.tools.transform import GeneralTransformer, RandomTransform
-from autots.tools.shaping import long_to_wide
+from autots.tools.shaping import long_to_wide, infer_frequency
 from autots.tools.regressor import create_lagged_regressor, create_regressor
-from autots.evaluator.auto_model import model_forecast
+from autots.evaluator.auto_model import model_forecast, ModelPrediction
 from autots.evaluator.anomaly_detector import AnomalyDetector, HolidayDetector
 from autots.models.cassandra import Cassandra
 
 
-__version__ = '0.5.6'
+__version__ = '0.6.18'
 
 TransformTS = GeneralTransformer
 
@@ -53,4 +54,6 @@ __all__ = [
     'AnomalyDetector',
     'HolidayDetector',
     'Cassandra',
+    'infer_frequency',
+    'ModelPrediction',
 ]
