@@ -11,9 +11,10 @@ from autots.models.model_list import all_result_path, diff_window_motif_list
 
 
 def extract_result_windows(forecasts, model_name=None):
+    """standardize result windows from different models."""
     result_windows = forecasts.model.result_windows
     if model_name is None:
-        model_name = forecasts.model.name
+        model_name = forecasts.model_name
     if model_name in diff_window_motif_list:
         result_windows = np.moveaxis(np.array(list(result_windows.values())), 0, -1)
     if result_windows.ndim == 4:
@@ -33,6 +34,16 @@ def extract_result_windows(forecasts, model_name=None):
     return np.array(transformed_array)
 
 
+def extract_window_index(forecasts):
+    model_name = forecasts.model_name
+    if model_name == "SectionalMotif":
+        return forecasts.model.windows
+    if model_name == "BallTreeMultivariateMotif":
+        return forecasts.model.windows
+    else:
+        raise ValueError("window indexes not supported by this model yet")
+
+
 def set_limit_forecast(
     df_train,
     forecast_length,
@@ -200,6 +211,7 @@ class EventRiskForecast(object):
             "verbose": 1,
             "n_jobs": "auto",
             "random_seed": 321,
+            "transformer_list": "no_expanding",
         },
         future_regressor_train=None,
         future_regressor_forecast=None,
@@ -272,9 +284,9 @@ class EventRiskForecast(object):
         if isinstance(prediction_interval, list):
             prediction_interval = prediction_interval[0]
         model = AutoTS(
-            forecast_length=self.forecast_length
-            if forecast_length is None
-            else forecast_length,
+            forecast_length=(
+                self.forecast_length if forecast_length is None else forecast_length
+            ),
             prediction_interval=prediction_interval,
             models_mode=models_mode,
             model_list=model_list,
