@@ -3,9 +3,12 @@ GluonTS
 
 Best neuralnet models currently available, released by Amazon, scale well.
 Except it is really the only thing I use that runs mxnet, and it takes a while to train these guys...
+And MXNet is now sorta-maybe-deprecated? Which is sad because it had excellent CPU-based training speed.
 
-Note that there are routinely package version issues with this and its dependencies. Stability is not the strong suit of GluonTS.
+Note that there are routinely package version issues with this and its dependencies.
+Stability is not the strong suit of GluonTS.
 """
+
 import logging
 import random
 import datetime
@@ -21,10 +24,13 @@ try:
         from gluonts.dataset.field_names import FieldName  # new way
     except Exception:
         from gluonts.transform import FieldName  # old way (0.3.3 and older)
-    try:  # new way
-        from gluonts.mx.trainer import Trainer
-    except Exception:  # old way < 0.5.x
-        from gluonts.trainer import Trainer
+    try:
+        try:  # new way, but only with mxnet
+            from gluonts.mx.trainer import Trainer
+        except Exception:  # old way < 0.5.x
+            from gluonts.trainer import Trainer
+    except Exception:
+        pass
 except Exception:  # except ImportError
     _has_gluonts = False
 else:
@@ -93,11 +99,9 @@ class GluonTS(ModelObject):
         """
         if not _has_gluonts:
             raise ImportError(
-                "GluonTS installation is incompatible with AutoTS. The numpy version is sometimes the issue."
+                "GluonTS installation is incompatible with AutoTS. The numpy version is sometimes the issue, try 1.23.1 {as of 06-2023}"
             )
 
-        df = self.basic_profile(df)
-
         try:
             from mxnet.random import seed as mxnet_seed
 
@@ -105,16 +109,10 @@ class GluonTS(ModelObject):
         except Exception:
             pass
 
-        gluon_train = df.to_numpy().T
+        gluon_freq = str(self.frequency).split('-')[0]
         self.train_index = df.columns
         self.train_columns = df.index
 
-        gluon_freq = str(self.frequency).split('-')[0]
-        if self.regression_type == "User":
-            if future_regressor is None:
-                raise ValueError(
-                    "regression_type='User' but no future_regressor supplied"
-                )
         if gluon_freq in ["MS", "1MS"]:
             gluon_freq = "M"
 
@@ -135,7 +133,7 @@ class GluonTS(ModelObject):
         else:
             self.gluon_context_length = 20
             self.context_length = '20'
-        ts_metadata = {
+        self.ts_metadata = ts_metadata = {
             'num_series': len(self.train_index),
             'freq': gluon_freq,
             'start_ts': df.index[0],
@@ -145,68 +143,61 @@ class GluonTS(ModelObject):
             'context_length': self.gluon_context_length,
             'forecast_length': self.forecast_length,
         }
+        self.fit_data(df, future_regressor=future_regressor)
         npts_flag = False
-        if self.gluon_model in self.multivariate_mods:
-            if self.regression_type == "User":
-                regr = future_regressor.to_numpy().T
-                self.regr_train = regr
-                self.test_ds = ListDataset(
-                    [
-                        {
-                            "start": df.index[0],
-                            "target": gluon_train,
-                            "feat_dynamic_real": regr,
-                        }
-                    ],
-                    freq=ts_metadata['freq'],
-                    one_dim_target=False,
-                )
-            else:
-                self.test_ds = ListDataset(
-                    [{"start": df.index[0], "target": gluon_train}],
-                    freq=ts_metadata['freq'],
-                    one_dim_target=False,
-                )
-        else:
-            if self.regression_type == "User":
-                self.gluon_train = gluon_train
-                regr = future_regressor.to_numpy().T
-                self.regr_train = regr
-                self.test_ds = ListDataset(
-                    [
-                        {
-                            FieldName.TARGET: target,
-                            FieldName.START: ts_metadata['start_ts'],
-                            FieldName.FEAT_DYNAMIC_REAL: regr,
-                        }
-                        for target in gluon_train
-                    ],
-                    freq=ts_metadata['freq'],
-                )
-            else:
-                # use the actual start date, if NaN given (semi-hidden)
-                # ts_metadata['gluon_start'] = df.notna().idxmax().tolist()
-                self.test_ds = ListDataset(
-                    [
-                        {FieldName.TARGET: target, FieldName.START: start}
-                        for (target, start) in zip(
-                            gluon_train, ts_metadata['gluon_start']
-                        )
-                    ],
-                    freq=ts_metadata['freq'],
+
+        pytorch_models = ['PatchTST', 'DeepAR']  # those supporting
+        if self.gluon_model in pytorch_models:
+            pass
+        """
+        # this attempts to stop model checkpoing saving spam that is typical of lightning
+            try:
+                try:
+                    from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
+                except Exception:
+                    from lightning.pytorch.callbacks.model_checkpoint import ModelCheckpoint
+    
+                callbacks = []
+                callbacks.append(
+                    ModelCheckpoint(
+                        save_last=False,
+                        save_top_k=0,
+                        every_n_train_steps=0,
+                        every_n_epochs=0,
+                    )
                 )
+            except Exception as e:
+                callbacks = []
+                print(repr(e))
+        """
+
         if self.gluon_model == 'DeepAR':
             try:
-                from gluonts.mx import DeepAREstimator
+                try:
+                    from gluonts.mx import DeepAREstimator
+                except Exception:
+                    from gluonts.model.deepar import DeepAREstimator
+                estimator = DeepAREstimator(
+                    freq=ts_metadata['freq'],
+                    context_length=ts_metadata['context_length'],
+                    prediction_length=ts_metadata['forecast_length'],
+                    trainer=Trainer(
+                        epochs=self.epochs, learning_rate=self.learning_rate
+                    ),
+                )
             except Exception:
-                from gluonts.model.deepar import DeepAREstimator
+                from gluonts.torch import DeepAREstimator
+
+                estimator = DeepAREstimator(
+                    freq=ts_metadata['freq'],
+                    context_length=ts_metadata['context_length'],
+                    prediction_length=ts_metadata['forecast_length'],
+                    trainer_kwargs={
+                        'logger': False,
+                        'log_every_n_steps': 0,
+                    },  # , 'callbacks': callbacks
+                )
 
-            estimator = DeepAREstimator(
-                freq=ts_metadata['freq'],
-                context_length=ts_metadata['context_length'],
-                prediction_length=ts_metadata['forecast_length'],
-                trainer=Trainer(epochs=self.epochs, learning_rate=self.learning_rate),
-            )
         elif self.gluon_model == 'NPTS':
             try:
                 from gluonts.model.npts import NPTSPredictor
@@ -319,7 +310,7 @@ class GluonTS(ModelObject):
                 from gluonts.model.deepvar import DeepVAREstimator
 
             estimator = DeepVAREstimator(
-                target_dim=gluon_train.shape[0],
+                target_dim=df.shape[1],
                 freq=ts_metadata['freq'],
                 context_length=ts_metadata['context_length'],
                 prediction_length=ts_metadata['forecast_length'],
@@ -332,7 +323,7 @@ class GluonTS(ModelObject):
                 from gluonts.model.gpvar import GPVAREstimator
 
             estimator = GPVAREstimator(
-                target_dim=gluon_train.shape[0],
+                target_dim=df.shape[1],
                 freq=ts_metadata['freq'],
                 context_length=ts_metadata['context_length'],
                 prediction_length=ts_metadata['forecast_length'],
@@ -436,6 +427,16 @@ class GluonTS(ModelObject):
                     hybridize=False,
                 ),
             )
+        elif self.gluon_model == 'PatchTST':
+            from gluonts.torch.model.patch_tst import PatchTSTEstimator
+
+            estimator = PatchTSTEstimator(
+                prediction_length=ts_metadata['forecast_length'],
+                context_length=ts_metadata['context_length'],
+                patch_len=5,
+                lr=self.learning_rate,
+                trainer_kwargs={'logger': False, 'log_every_n_steps': 0},
+            )
         else:
             raise ValueError("'gluon_model' not recognized.")
 
@@ -443,12 +444,75 @@ class GluonTS(ModelObject):
             self.GluonPredictor = estimator
         else:
             self.GluonPredictor = estimator.train(self.test_ds)
-        self.ts_metadata = ts_metadata
         self.fit_runtime = datetime.datetime.now() - self.startTime
         return self
 
+    def fit_data(self, df, future_regressor=None):
+        df = self.basic_profile(df)
+        gluon_train = df.to_numpy().T
+        self.train_index = df.columns
+        self.train_columns = df.index
+        if self.regression_type == "User":
+            if future_regressor is None:
+                raise ValueError(
+                    "regression_type='User' but no future_regressor supplied"
+                )
+        if self.gluon_model in self.multivariate_mods:
+            if self.regression_type == "User":
+                regr = future_regressor.to_numpy().T
+                self.regr_train = regr
+                self.test_ds = ListDataset(
+                    [
+                        {
+                            "start": df.index[0],
+                            "target": gluon_train,
+                            "feat_dynamic_real": regr,
+                        }
+                    ],
+                    freq=self.ts_metadata['freq'],
+                    one_dim_target=False,
+                )
+            else:
+                self.test_ds = ListDataset(
+                    [{"start": df.index[0], "target": gluon_train}],
+                    freq=self.ts_metadata['freq'],
+                    one_dim_target=False,
+                )
+        else:
+            if self.regression_type == "User":
+                self.gluon_train = gluon_train
+                regr = future_regressor.to_numpy().T
+                self.regr_train = regr
+                self.test_ds = ListDataset(
+                    [
+                        {
+                            FieldName.TARGET: target,
+                            FieldName.START: self.ts_metadata['start_ts'],
+                            FieldName.FEAT_DYNAMIC_REAL: regr,
+                        }
+                        for target in gluon_train
+                    ],
+                    freq=self.ts_metadata['freq'],
+                )
+            else:
+                # use the actual start date, if NaN given (semi-hidden)
+                # ts_metadata['gluon_start'] = df.notna().idxmax().tolist()
+                self.test_ds = ListDataset(
+                    [
+                        {FieldName.TARGET: target, FieldName.START: start}
+                        for (target, start) in zip(
+                            gluon_train, self.ts_metadata['gluon_start']
+                        )
+                    ],
+                    freq=self.ts_metadata['freq'],
+                )
+        return self
+
     def predict(
-        self, forecast_length: int, future_regressor=[], just_point_forecast=False
+        self,
+        forecast_length: int = None,
+        future_regressor=[],
+        just_point_forecast=False,
     ):
         """Generates forecast data immediately following dates of index supplied to .fit()
 
@@ -461,6 +525,8 @@ class GluonTS(ModelObject):
             Either a PredictionObject of forecasts and metadata, or
             if just_point_forecast == True, a dataframe of point forecasts
         """
+        if forecast_length is None:
+            forecast_length = self.forecast_length
         if int(forecast_length) > int(self.forecast_length):
             raise ValueError("GluonTS must be refit to change forecast length!")
         predictStartTime = datetime.datetime.now()
@@ -605,7 +671,7 @@ class GluonTS(ModelObject):
                     0.1,
                     0.1,
                     0.05,
-                    0.1,
+                    0.01,
                     0.1,
                     0.1,
                     0.01,
@@ -637,6 +703,7 @@ class GluonTS(ModelObject):
                     'SelfAttention',
                     'TemporalFusionTransformer',
                     'DeepTPP',
+                    'PatchTST',
                 ],
                 [
                     0.1,
@@ -656,6 +723,7 @@ class GluonTS(ModelObject):
                     0.1,
                     0.1,
                     0.1,
+                    0.1,
                 ],
                 k=1,
             )[0]
@@ -667,7 +735,9 @@ class GluonTS(ModelObject):
             [5, 10, 30, '1ForecastLength', '2ForecastLength'],
             [0.2, 0.3, 0.1, 0.1, 0.1],
         )[0]
-        learning_rate_choice = random.choices([0.01, 0.001, 0.0001], [0.3, 0.6, 0.1])[0]
+        learning_rate_choice = random.choices(
+            [0.01, 0.001, 0.0001, 0.00001], [0.3, 0.6, 0.1, 0.1]
+        )[0]
         # NPTS doesn't use these, so just fill a constant
         if gluon_model_choice in ['NPTS', 'Rotbaum']:
             epochs_choice = 20
@@ -706,7 +776,7 @@ class GluonTS(ModelObject):
 
 """
 model = GluonTS(epochs=5)
-model = model.fit(df.fillna(method='ffill').fillna(method='bfill'))
+model = model.fit(df.ffill().bfill())
 prediction = model.predict(forecast_length=14)
 prediction.forecast
 """
