@@ -4,6 +4,7 @@ Statsmodels documentation can be a bit confusing.
 And it seems standard at first, but each model likes to do things differently.
 For example: exog, exog_oos, and exog_fc all sometimes mean the same thing
 """
+
 import datetime
 import warnings
 import random
@@ -11,7 +12,14 @@ import numpy as np
 import pandas as pd
 from autots.models.base import ModelObject, PredictionObject
 from autots.tools.probabilistic import Point_to_Probability
-from autots.tools.seasonal import date_part, seasonal_int, date_part_methods
+from autots.tools.seasonal import (
+    date_part,
+    seasonal_int,
+    date_part_methods,
+    base_seasonalities,
+    create_changepoint_features,
+    changepoint_fcst_from_last_row,
+)
 from autots.tools.holiday import holiday_flag
 
 # these are optional packages
@@ -45,6 +53,9 @@ class GLS(ModelObject):
         prediction_interval: float = 0.9,
         holiday_country: str = 'US',
         random_seed: int = 2020,
+        changepoint_spacing: int = None,
+        changepoint_distance_end: int = None,
+        constant: bool = False,
         **kwargs,
     ):
         ModelObject.__init__(
@@ -55,6 +66,9 @@ class GLS(ModelObject):
             holiday_country=holiday_country,
             random_seed=random_seed,
         )
+        self.changepoint_spacing = changepoint_spacing
+        self.changepoint_distance_end = changepoint_distance_end
+        self.constant = constant
 
     def fit(self, df, future_regressor=None):
         """Train algorithm given data supplied
@@ -67,6 +81,26 @@ class GLS(ModelObject):
         df = self.basic_profile(df)
         self.df_train = df
         Xf = pd.to_numeric(df.index, errors='coerce', downcast='integer').values
+        if self.constant:
+            Xf = np.concatenate(
+                (Xf.reshape(-1, 1), np.ones_like(Xf).reshape(-1, 1)), axis=1
+            )
+        if self.changepoint_spacing is not None:
+            if self.changepoint_distance_end is None:
+                changepoint_distance_end = self.changepoint_spacing
+            else:
+                changepoint_distance_end = self.changepoint_distance_end
+            x_t = create_changepoint_features(
+                self.df_train.index,
+                changepoint_spacing=self.changepoint_spacing,
+                changepoint_distance_end=changepoint_distance_end,
+            )
+            self.last_row = x_t.iloc[-1]
+            if Xf.ndim == 1:
+                Xf = np.concatenate((Xf.reshape(-1, 1), x_t.to_numpy()), axis=1)
+            else:
+                Xf = np.concatenate((Xf, x_t.to_numpy()), axis=1)
+        self.X = Xf
         self.model = GLS(df.values, Xf, missing='drop').fit()
         self.fit_runtime = datetime.datetime.now() - self.startTime
         return self
@@ -88,6 +122,18 @@ class GLS(ModelObject):
         predictStartTime = datetime.datetime.now()
         index = self.create_forecast_index(forecast_length=forecast_length)
         Xf = pd.to_numeric(index, errors='coerce', downcast='integer').values
+        if self.constant:
+            Xf = np.concatenate(
+                (Xf.reshape(-1, 1), np.ones_like(Xf).reshape(-1, 1)), axis=1
+            )
+        if self.changepoint_spacing is not None:
+            # this a funny little hack for making a dataframe from a row
+            x_tf = changepoint_fcst_from_last_row(self.last_row, int(forecast_length))
+            if Xf.ndim == 1:
+                Xf = np.concatenate((Xf.reshape(-1, 1), x_tf.to_numpy()), axis=1)
+            else:
+                Xf = np.concatenate((Xf, x_tf.to_numpy()), axis=1)
+        self.Xf = Xf
         forecast = self.model.predict(Xf)
         df = pd.DataFrame(forecast, columns=self.column_names, index=index)
         if just_point_forecast:
@@ -119,11 +165,25 @@ class GLS(ModelObject):
 
     def get_new_params(self, method: str = 'random'):
         """Returns dict of new parameters for parameter tuning"""
-        return {}
+        return {
+            "changepoint_spacing": random.choices(
+                [None, 6, 28, 60, 90, 180, 360, 5040],
+                [0.6, 0.05, 0.1, 0.1, 0.1, 0.2, 0.1, 0.2],
+            )[0],
+            "changepoint_distance_end": random.choices(
+                [None, 6, 28, 60, 90, 180, 360, 5040],
+                [0.6, 0.1, 0.1, 0.1, 0.1, 0.2, 0.1, 0.2],
+            )[0],
+            "constant": random.choices([True, False], [0.5, 0.5])[0],
+        }
 
     def get_params(self):
         """Return dict of current parameters"""
-        return {}
+        return {
+            "changepoint_spacing": self.changepoint_spacing,
+            "changepoint_distance_end": self.changepoint_distance_end,
+            "constant": self.constant,
+        }
 
 
 def glm_forecast_by_column(current_series, X, Xf, args):
@@ -136,36 +196,36 @@ def glm_forecast_by_column(current_series, X, Xf, args):
     if str(family).lower() == 'poisson':
         from statsmodels.genmod.families.family import Poisson
 
-        model = SM_GLM(current_series.values, X, family=Poisson(), missing='drop').fit(
-            disp=verbose
-        )
+        model = SM_GLM(
+            current_series.to_numpy(), X, family=Poisson(), missing='drop'
+        ).fit(disp=verbose)
     elif str(family).lower() == 'binomial':
         from statsmodels.genmod.families.family import Binomial
 
-        model = SM_GLM(current_series.values, X, family=Binomial(), missing='drop').fit(
-            disp=verbose
-        )
+        model = SM_GLM(
+            current_series.to_numpy(), X, family=Binomial(), missing='drop'
+        ).fit(disp=verbose)
     elif str(family).lower() == 'negativebinomial':
         from statsmodels.genmod.families.family import NegativeBinomial
 
         model = SM_GLM(
-            current_series.values, X, family=NegativeBinomial(), missing='drop'
+            current_series.to_numpy(), X, family=NegativeBinomial(), missing='drop'
         ).fit(disp=verbose)
     elif str(family).lower() == 'tweedie':
         from statsmodels.genmod.families.family import Tweedie
 
-        model = SM_GLM(current_series.values, X, family=Tweedie(), missing='drop').fit(
-            disp=verbose
-        )
+        model = SM_GLM(
+            current_series.to_numpy(), X, family=Tweedie(), missing='drop'
+        ).fit(disp=verbose)
     elif str(family).lower() == 'gamma':
         from statsmodels.genmod.families.family import Gamma
 
-        model = SM_GLM(current_series.values, X, family=Gamma(), missing='drop').fit(
-            disp=verbose
-        )
+        model = SM_GLM(
+            current_series.to_numpy(), X, family=Gamma(), missing='drop'
+        ).fit(disp=verbose)
     else:
         family = 'Gaussian'
-        model = SM_GLM(current_series.values, X, missing='drop').fit(disp=verbose)
+        model = SM_GLM(current_series.to_numpy(), X, missing='drop').fit(disp=verbose)
     Pred = model.predict((Xf))
     Pred = pd.Series(Pred)
     Pred.name = series_name
@@ -193,6 +253,7 @@ class GLM(ModelObject):
         regression_type: str = None,
         family='Gaussian',
         constant: bool = False,
+        changepoint_spacing: int = None,
         verbose: int = 1,
         n_jobs: int = None,
         **kwargs,
@@ -210,6 +271,7 @@ class GLM(ModelObject):
         )
         self.family = family
         self.constant = constant
+        self.changepoint_spacing = changepoint_spacing
 
     def fit(self, df, future_regressor=None):
         """Train algorithm given data supplied
@@ -250,11 +312,25 @@ class GLM(ModelObject):
         test_index = self.create_forecast_index(forecast_length=forecast_length)
 
         if self.regression_type == 'datepart':
-            X = date_part(self.df_train.index, method='expanded').values
+            X = date_part(self.df_train.index, method='expanded').to_numpy()
+        elif self.regression_type in base_seasonalities or isinstance(
+            self.regression_type, list
+        ):
+            X = date_part(self.df_train.index, method=self.regression_type).to_numpy()
         else:
             X = pd.to_numeric(
                 self.df_train.index, errors='coerce', downcast='integer'
-            ).values
+            ).to_numpy()
+        if self.changepoint_spacing is not None:
+            x_t = create_changepoint_features(
+                self.df_train.index,
+                changepoint_spacing=self.changepoint_spacing,
+                changepoint_distance_end=self.changepoint_spacing,
+            )
+            if X.ndim == 1:
+                X = np.concatenate((X.reshape(-1, 1), x_t.to_numpy()), axis=1)
+            else:
+                X = np.concatenate((X, x_t.to_numpy()), axis=1)
         if self.constant in [True, 'True', 'true']:
             from statsmodels.tools import add_constant
 
@@ -270,9 +346,23 @@ class GLM(ModelObject):
         fill_vals = self.df_train.abs().min(axis=0, skipna=True)
         self.df_train = self.df_train.fillna(fill_vals).fillna(0.1)
         if self.regression_type == 'datepart':
-            Xf = date_part(test_index, method='expanded').values
+            Xf = date_part(test_index, method='expanded').to_numpy()
+        elif self.regression_type in base_seasonalities or isinstance(
+            self.regression_type, list
+        ):
+            Xf = date_part(test_index, method=self.regression_type).to_numpy()
         else:
-            Xf = pd.to_numeric(test_index, errors='coerce', downcast='integer').values
+            Xf = pd.to_numeric(
+                test_index, errors='coerce', downcast='integer'
+            ).to_numpy()
+
+        if self.changepoint_spacing is not None:
+            # this a funny little hack for making a dataframe from a row
+            x_tf = changepoint_fcst_from_last_row(x_t.iloc[-1], int(forecast_length))
+            if Xf.ndim == 1:
+                Xf = np.concatenate((Xf.reshape(-1, 1), x_tf.to_numpy()), axis=1)
+            else:
+                Xf = np.concatenate((Xf, x_tf.to_numpy()), axis=1)
         if self.constant or self.constant == 'True':
             Xf = add_constant(Xf, has_constant='add')
         if self.regression_type == 'User':
@@ -298,7 +388,7 @@ class GLM(ModelObject):
             parallel = False
         # joblib multiprocessing to loop through series
         if parallel:
-            df_list = Parallel(n_jobs=self.n_jobs, verbose=pool_verbose)(
+            df_list = Parallel(n_jobs=self.n_jobs, verbose=pool_verbose, timeout=7200)(
                 delayed(glm_forecast_by_column)(
                     current_series=df[col],
                     X=X,
@@ -362,10 +452,15 @@ class GLM(ModelObject):
             regression_type_choice = random.choices(
                 [None, 'datepart', 'User'], [0.4, 0.4, 0.2]
             )[0]
+            if regression_type_choice == "datepart":
+                regression_type_choice = random.choice(base_seasonalities)
         return {
             'family': family_choice,
             'constant': constant_choice,
             'regression_type': regression_type_choice,
+            'changepoint_spacing': random.choices(
+                [None, 30, 60, 180, 5040], [0.5, 0.1, 0.1, 0.1, 0.1]
+            )[0],
         }
 
     def get_params(self):
@@ -374,6 +469,7 @@ class GLM(ModelObject):
             'family': self.family,
             'constant': self.constant,
             'regression_type': self.regression_type,
+            'changepoint_spacing': self.changepoint_spacing,
         }
 
 
@@ -400,6 +496,7 @@ class ETS(ModelObject):
         trend: str = None,
         seasonal: str = None,
         seasonal_periods: int = None,
+        method: str = None,
         holiday_country: str = 'US',
         random_seed: int = 2020,
         verbose: int = 0,
@@ -426,6 +523,7 @@ class ETS(ModelObject):
             self.seasonal_periods = None
         else:
             self.seasonal_periods = abs(int(seasonal_periods))
+        self.method = method
 
     def fit(self, df, future_regressor=None):
         """Train algorithm given data supplied
@@ -454,7 +552,6 @@ class ETS(ModelObject):
             if just_point_forecast == True, a dataframe of point forecasts
         """
         predictStartTime = datetime.datetime.now()
-        from statsmodels.tsa.holtwinters import ExponentialSmoothing
 
         test_index = self.create_forecast_index(forecast_length=forecast_length)
         parallel = True
@@ -466,9 +563,19 @@ class ETS(ModelObject):
             'freq': self.frequency,
             'forecast_length': forecast_length,
             'verbose': self.verbose,
+            "method": self.method,
         }
 
-        def ets_forecast_by_column(current_series, args):
+        cols = self.df_train.columns.tolist()
+        if self.n_jobs in [0, 1] or len(cols) < 4:
+            parallel = False
+        elif not joblib_present:
+            parallel = False
+
+        # seems to perhaps be more stable when defined inside, unclear
+        from statsmodels.tsa.holtwinters import ExponentialSmoothing
+
+        def ets_forecast_by_column(current_series, args, test_index):
             """Run one series of ETS and return prediction."""
             series_name = current_series.name
             with warnings.catch_warnings():
@@ -487,7 +594,7 @@ class ETS(ModelObject):
                             # freq=args['freq'],
                         )
                     except Exception as e:
-                        if args['verbose'] > 0:
+                        if args['verbose'] > 1:
                             print(f"ETS error {repr(e)}")
                         esModel = ExponentialSmoothing(
                             current_series,
@@ -498,7 +605,7 @@ class ETS(ModelObject):
                             # initialization_method='heuristic',  # estimated
                             freq=args['freq'],
                         )
-                    esResult = esModel.fit()
+                    esResult = esModel.fit(method=args["method"])
                     srt = current_series.shape[0]
                     esPred = esResult.predict(
                         start=srt, end=srt + args['forecast_length'] - 1
@@ -506,28 +613,29 @@ class ETS(ModelObject):
                     esPred = pd.Series(esPred)
                 except Exception as e:
                     # this error handling is meant for horizontal ensembles where it will only then be needed for select series
-                    if args['verbose'] > 0:
+                    if args['verbose'] > 1:
                         print(f"ETS failed on {series_name} with {repr(e)}")
-                    esPred = pd.Series((np.zeros((forecast_length,))), index=test_index)
+                    esPred = pd.Series(
+                        (np.zeros((args["forecast_length"],))), index=test_index
+                    )
             esPred.name = series_name
             return esPred
 
-        cols = self.df_train.columns.tolist()
-        if self.n_jobs in [0, 1] or len(cols) < 4:
-            parallel = False
-        elif not joblib_present:
-            parallel = False
         # joblib multiprocessing to loop through series
         if parallel:
-            df_list = Parallel(n_jobs=self.n_jobs)(
-                delayed(ets_forecast_by_column)(self.df_train[col], args)
-                for (col) in cols
+            df_list = Parallel(n_jobs=self.n_jobs, timeout=36000)(
+                delayed(ets_forecast_by_column)(
+                    self.df_train[col].astype(float), args, test_index
+                )
+                for col in cols
             )
             forecast = pd.concat(df_list, axis=1)
         else:
             df_list = []
             for col in cols:
-                df_list.append(ets_forecast_by_column(self.df_train[col], args))
+                df_list.append(
+                    ets_forecast_by_column(self.df_train[col], args, test_index)
+                )
             forecast = pd.concat(df_list, axis=1)
         if just_point_forecast:
             return forecast
@@ -569,7 +677,9 @@ class ETS(ModelObject):
         seasonal_probability = [0.2, 0.2, 0.6]
         seasonal_choice = random.choices(seasonal_list, seasonal_probability)[0]
         if seasonal_choice in ["additive", "multiplicative"]:
-            seasonal_period_choice = seasonal_int()
+            seasonal_period_choice = seasonal_int(
+                small=True if method != "deep" else False
+            )
         else:
             seasonal_period_choice = None
         parameter_dict = {
@@ -577,6 +687,9 @@ class ETS(ModelObject):
             'trend': trend_choice,
             'seasonal': seasonal_choice,
             'seasonal_periods': seasonal_period_choice,
+            'method': random.choices(
+                [None, "SLSQP", "L-BFGS-B", "least_squares"], [0.3, 0.2, 0.2, 0.2]
+            )[0],
         }
         return parameter_dict
 
@@ -587,6 +700,7 @@ class ETS(ModelObject):
             'trend': self.trend,
             'seasonal': self.seasonal,
             'seasonal_periods': self.seasonal_periods,
+            'method': self.method,
         }
         return parameter_dict
 
@@ -756,7 +870,7 @@ class ARIMA(ModelObject):
         # joblib multiprocessing to loop through series
         if parallel:
             verbs = 0 if self.verbose < 1 else self.verbose - 1
-            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs))(
+            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs), timeout=36000)(
                 delayed(arima_seek_the_oracle)(
                     current_series=self.df_train[col], args=args, series=col
                 )
@@ -1003,7 +1117,7 @@ class UnobservedComponents(ModelObject):
                         # damped_cycle=args['damped_cycle'],
                         # irregular=args['irregular'],
                         autoregressive=args['autoregressive'],
-                        **args['model_kwargs']
+                        **args['model_kwargs'],
                         # stochastic_cycle=args['stochastic_cycle'],
                         # stochastic_level=args['stochastic_level'],
                         # stochastic_trend=args['stochastic_trend'],
@@ -1040,7 +1154,7 @@ class UnobservedComponents(ModelObject):
         # joblib multiprocessing to loop through series
         if parallel:
             verbs = 0 if self.verbose < 1 else self.verbose - 1
-            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs))(
+            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs), timeout=36000)(
                 delayed(uc_forecast_by_column)(
                     current_series=self.df_train[col],
                     args=args,
@@ -1136,7 +1250,8 @@ class UnobservedComponents(ModelObject):
                 ["lbfgs", "bfgs", "powell", "cg", "newton", "nm"],
                 [0.8, 0.1, 0.1, 0.1, 0.1, 0.1],
             )[0],
-            'autoregressive': random.choices([None, 1, 2], [0.8, 0.2, 0.1])[0],
+            # AR1 is helpful but AR2 is getting slow
+            'autoregressive': random.choices([None, 1, 2], [0.8, 0.3, 0.02])[0],
             'regression_type': regression_choice,
         }
 
@@ -1363,6 +1478,8 @@ class VECM(ModelObject):
         verbose: int = 0,
         deterministic: str = 'n',
         k_ar_diff: int = 1,
+        seasons: int = 0,
+        coint_rank: int = 1,
         **kwargs,
     ):
         ModelObject.__init__(
@@ -1377,6 +1494,8 @@ class VECM(ModelObject):
         )
         self.deterministic = deterministic
         self.k_ar_diff = k_ar_diff
+        self.seasons = seasons
+        self.coint_rank = coint_rank
 
     def fit(self, df, future_regressor=None):
         """Train algorithm given data supplied.
@@ -1427,21 +1546,24 @@ class VECM(ModelObject):
             ), "regressor row count not equal to forecast length"
 
         # LinAlgError: SVD did not converge (occurs when NaN in train data)
+        # NaN must be removed for some BLAS packages else they will kill the kernel
         if self.regression_type in ["User", "Holiday", 'user']:
             maModel = VECM(
-                self.df_train,
+                self.df_train.replace([np.inf, -np.inf], np.nan).fillna(0),
                 freq=self.frequency,
-                exog=np.array(self.regressor_train),
+                exog=np.nan_to_num(np.array(self.regressor_train)),
                 deterministic=self.deterministic,
                 k_ar_diff=self.k_ar_diff,
+                coint_rank=self.coint_rank,
+                seasons=self.seasons,
             ).fit()
             # don't ask me why it is exog_fc here and not exog like elsewhere
             forecast = maModel.predict(
-                steps=forecast_length, exog_fc=np.array(future_regressor)
+                steps=forecast_length, exog_fc=np.nan_to_num(np.array(future_regressor))
             )
         else:
             maModel = VECM(
-                self.df_train,
+                self.df_train.replace([np.inf, -np.inf], np.nan).fillna(0),
                 freq=self.frequency,
                 deterministic=self.deterministic,
                 k_ar_diff=self.k_ar_diff,
@@ -1478,30 +1600,28 @@ class VECM(ModelObject):
 
     def get_new_params(self, method: str = 'random'):
         """Return dict of new parameters for parameter tuning."""
-        deterministic_choice = np.random.choice(
-            a=["n", "co", "ci", "lo", "li", "cili", "colo"],
-            size=1,
-            p=[0.4, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
-        ).item()
-        k_ar_diff_choice = np.random.choice(
-            a=[0, 1, 2, 3], size=1, p=[0.1, 0.5, 0.2, 0.2]
-        ).item()
+        deterministic_choice = random.choices(
+            ["n", "co", "ci", "lo", "li", "cili", "colo"],
+            [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
+        )[0]
+        k_ar_diff_choice = random.choices([0, 1, 2, 3], [0.1, 0.5, 0.2, 0.2])[0]
 
         if "regressor" in method:
             regression_choice = "User"
         else:
             regression_list = [None, 'User', 'Holiday']
             regression_probability = [0.8, 0.15, 0.05]
-            regression_choice = np.random.choice(
-                a=regression_list, size=1, p=regression_probability
-            ).item()
+            regression_choice = random.choices(regression_list, regression_probability)[
+                0
+            ]
 
-        parameter_dict = {
+        return {
             'deterministic': deterministic_choice,
             'k_ar_diff': k_ar_diff_choice,
+            'seasons': random.choices([0, 7, 12], [0.9, 0.1, 0.1])[0],
+            'coint_rank': random.choices([1, 2, 3], [0.6, 0.2, 0.2])[0],
             'regression_type': regression_choice,
         }
-        return parameter_dict
 
     def get_params(self):
         """Return dict of current parameters."""
@@ -1959,7 +2079,7 @@ class Theta(ModelObject):
         # joblib multiprocessing to loop through series
         if parallel:
             verbs = 0 if self.verbose < 1 else self.verbose - 1
-            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs))(
+            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs), timeout=36000)(
                 delayed(theta_forecast_by_column)(
                     current_series=self.df_train[col], args=args
                 )
@@ -2000,14 +2120,18 @@ class Theta(ModelObject):
 
     def get_new_params(self, method: str = 'random'):
         """Return dict of new parameters for parameter tuning."""
+        if method in ["deep"]:
+            period = random.choices([None, 7, 28, 288], [0.8, 0.1, 0.1, 0.01])[0]
+        else:
+            period = None
         return {
             'deseasonalize': random.choices([True, False], [0.8, 0.2])[0],
             'difference': random.choice([True, False]),
-            'use_test': random.choices([True, False], [0.8, 0.2])[0],
+            'use_test': random.choices([True, False], [0.4, 0.2])[0],
             'method': "auto",
-            'period': None,
+            'period': period,
             'theta': random.choice([1.2, 1.4, 1.6, 2, 2.5, 3, 4]),
-            'use_mle': random.choices([True, False], [0.2, 0.8])[0],
+            'use_mle': random.choices([True, False], [0.0001, 0.99])[0],
         }
 
     def get_params(self):
@@ -2201,7 +2325,7 @@ class ARDL(ModelObject):
         # joblib multiprocessing to loop through series
         if parallel:
             verbs = 0 if self.verbose < 1 else self.verbose - 1
-            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs))(
+            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs), timeout=72000)(
                 delayed(ardl_per_column)(
                     current_series=self.df_train[col],
                     args=args,
